//+------------------------------------------------------------------+
//|                                                          GHE.mqh |
//|                                  Copyright 2024, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#include<VectorMatrixTools.mqh>
//+------------------------------------------------------------------+
//| Calculates the generalized Hurst exponent H(q) from the scaling  |
//| of the renormalized q-moments of the distribution                |
//+------------------------------------------------------------------+
double general_hurst(double &data[], int q, int lower=2,int upper=50)
{
 if(data.Size()<100)
   {
    Print("data array is of insufficient length");
    return EMPTY_VALUE;
   }
 
 if( lower>=upper || lower<2 ||  upper>int(floor(0.5*data.Size())))
   {
    Print("Invalid input for lower and/or upper");
    return EMPTY_VALUE;
   }
 
 if(q<=0)
  {
   Print("Invalid input for q");
   return EMPTY_VALUE;
  }
    
 uint len = data.Size();
 
 int k =0;
 
  matrix H,mcord,lmcord;
  vector n_vector,dv,vv,Y,ddVd,VVVd,XP,XY,PddVd,PVVVd,Px_vector,Sqx,pt;
  double dv_array[],vv_array[],mx,SSxx,my,SSxy,cc1,cc2,N;
 
 if(!H.Resize(ulong(upper-lower),1))
    {
     Print(__LINE__," ",__FUNCTION__," ",GetLastError());
     return EMPTY_VALUE;
    }
     
 for(int i=lower;i<upper;i++)
  {
    vector x_vector(ulong(i),arange,1.0,1.0);
    
    if(!mcord.Resize(ulong(i),1))
       {
        Print(__LINE__," ",__FUNCTION__," ",GetLastError());
        return EMPTY_VALUE;
       }
     
    mcord.Fill(0.0);
 
    for(int j=1;j<i+1;j++)
     {
      if(!diff_array(j,data,dv,Y))
         return EMPTY_VALUE;
      
      N = double(Y.Size());
         
      vector X(ulong(N),arange,1.0,1.0);
      
      mx = X.Sum()/N;
      
      XP = MathPow(X,2.0);
      
      SSxx = XP.Sum() - N*pow(mx,2.0);
      
      my = Y.Sum()/N;
      
      XY = X*Y;
      
      SSxy = XY.Sum() - N*mx*my;
      
      cc1 = SSxy/SSxx;
      
      cc2 = my - cc1*mx;
      
      ddVd = dv - cc1;
      
      VVVd = Y - cc1*X - cc2;
      
      PddVd = MathAbs(ddVd);
      
      PddVd = pow(PddVd,q);
      
      PVVVd = MathAbs(VVVd);
      
      PVVVd = pow(PVVVd,q);
      
      mcord[j-1][0] = PddVd.Mean()/PVVVd.Mean();
     }   
   
   Px_vector = MathLog10(x_vector);
        
   mx = Px_vector.Mean();
   
   Sqx = MathPow(Px_vector,2.0);
   
   SSxx = Sqx.Sum() - i*pow(mx,2.0);
   
   lmcord = log10(mcord);
      
   my = lmcord.Mean();
   
   pt = Px_vector*lmcord.Col(0);
   
   SSxy = pt.Sum() - i*mx*my;
   
   H[k][0]= SSxy/SSxx;
   
   k++;
  }   
 
 //Print("H \n",H);
 
 return H.Mean()/double(q);

} 
//+--------------------------------------------------------------------------+
//|overloaded gen_hurst() function that works with series contained in vector|
//+--------------------------------------------------------------------------+
double general_hurst(vector &data, int q, int lower=0,int upper=0)
{
  double series[];
  
  if(!vecToArray(data,series))
    return EMPTY_VALUE;
  else
    return general_hurst(series,q,lower,upper);  
}
//+------------------------------------------------------------------+
//| Difference helper function                                       |
//+------------------------------------------------------------------+
bool diff_array(uint st,double &in[],vector &out,vector &other)
{
 ulong nsize = (MathMod(in.Size(),st)==0)?int(in.Size()/st)-1:int(in.Size()/st);
 
 if(!out.Resize(nsize) || !other.Resize(nsize+1))
   {
    Print("resize error ",GetLastError());
    return false;
   }
 
 uint i,tt=0;
 for(i = st; i<in.Size();i+=st,tt++)
     {
      other[tt]=in[i-st];
      out[tt]=in[i]-in[i-st];
     }
     
 other[tt]=in[i-st];
     
 return true;
}
